import itertools
from typing import Iterable, List, Set, Tuple

from grimp import DetailedImport, ImportGraph

from importlinter.domain.imports import (
    DirectImport,
    ImportExpression,
    Module,
    ModuleExpression,
)


class MissingImport(Exception):
    pass


def pop_imports(graph: ImportGraph, imports: Iterable[DirectImport]) -> List[DetailedImport]:
    """
    Removes the supplied direct imports from the graph.

    Returns:
        The list of import details that were removed, including any additional metadata.

    Raises:
        MissingImport if an import is not present in the graph.
    """
    removed_imports: List[DetailedImport] = []

    imports_to_remove = _dedupe_imports(imports)

    for import_to_remove in imports_to_remove:
        import_details = graph.get_import_details(
            importer=import_to_remove.importer.name,
            imported=import_to_remove.imported.name,
        )
        if not import_details:
            raise MissingImport(f"Ignored import {import_to_remove} not present in the graph.")

        graph.remove_import(
            importer=import_to_remove.importer.name,
            imported=import_to_remove.imported.name,
        )

        removed_imports.extend(import_details)

    return removed_imports


def import_expression_to_imports(
    graph: ImportGraph, expression: ImportExpression
) -> List[DirectImport]:
    """
    Returns a list of imports in a graph, given some import expression.

    Raises:
        MissingImport if an import is not present in the graph. For a wildcarded import expression,
        this is raised if there is not at least one match.
    """
    matching_imports = graph.find_matching_direct_imports(import_expression=str(expression))

    if not matching_imports:
        raise MissingImport(
            f"Ignored import expression {expression} didn't match anything in the graph."
        )

    detailed_imports: Set[DirectImport] = set()
    for matching_import in matching_imports:
        import_details = graph.get_import_details(
            importer=matching_import["importer"],
            imported=matching_import["imported"],
        )

        if import_details:
            for individual_import_details in import_details:
                detailed_imports.add(
                    DirectImport(
                        importer=Module(individual_import_details["importer"]),
                        imported=Module(individual_import_details["imported"]),
                        line_number=individual_import_details["line_number"],
                        line_contents=individual_import_details["line_contents"],
                    )
                )

    return list(detailed_imports)


def module_expressions_to_modules(
    graph: ImportGraph, expressions: Iterable[ModuleExpression]
) -> Set[Module]:
    modules = set()
    for expression in expressions:
        modules |= module_expression_to_modules(graph, expression)
    return modules


def module_expression_to_modules(graph: ImportGraph, expression: ModuleExpression) -> Set[Module]:
    if not expression.has_wildcard_expression():
        return {Module(expression.expression)}

    matching_modules = graph.find_matching_modules(expression.expression)
    return {Module(module) for module in matching_modules}


def import_expressions_to_imports(
    graph: ImportGraph, expressions: Iterable[ImportExpression]
) -> List[DirectImport]:
    """
    Returns a list of imports in a graph, given some import expressions.

    Raises:
        MissingImport if an import is not present in the graph. For a wildcarded import expression,
        this is raised if there is not at least one match.
    """
    return list(
        set(
            itertools.chain(
                *(import_expression_to_imports(graph, expression) for expression in expressions)
            )
        )
    )


def resolve_import_expressions(
    graph: ImportGraph, expressions: Iterable[ImportExpression]
) -> Tuple[Set[DirectImport], Set[ImportExpression]]:
    """
    Find any imports in the graph that match the supplied import expressions.

    Returns tuple of:
        - Set of resolved imports.
        - Set of import expressions that didn't match any imports.
    """
    resolved_imports = set()
    unresolved_expressions = set()

    for expression in expressions:
        try:
            resolved_imports.update(import_expression_to_imports(graph, expression))
        except MissingImport:
            unresolved_expressions.add(expression)

    return (resolved_imports, unresolved_expressions)


def pop_import_expressions(
    graph: ImportGraph, expressions: Iterable[ImportExpression]
) -> List[DetailedImport]:
    """
    Removes any imports matching the supplied import expressions from the graph.

    Returns:
        The list of imports that were removed, including any additional metadata.
    Raises:
        MissingImport if an import is not present in the graph. For a wildcarded import expression,
        this is raised if there is not at least one match.
    """
    imports = import_expressions_to_imports(graph, expressions)
    return pop_imports(graph, imports)


def add_imports(graph: ImportGraph, import_details: List[DetailedImport]) -> None:
    """
    Adds the supplied import details to the graph.

    Intended to be the reverse of pop_imports, so the following code should leave the
    graph unchanged:

        import_details = pop_imports(graph, imports)
        add_imports(graph, import_details)
    """
    for details in import_details:
        assert isinstance(details["importer"], str)
        assert isinstance(details["imported"], str)
        assert isinstance(details["line_number"], int)
        assert isinstance(details["line_contents"], str)
        graph.add_import(
            importer=details["importer"],
            imported=details["imported"],
            line_number=details["line_number"],
            line_contents=details["line_contents"],
        )


def _dedupe_imports(imports: Iterable[DirectImport]) -> Iterable[DirectImport]:
    """
    Return the imports with the metadata and any duplicates removed.

    For example:

        _dedupe_imports([
            DirectImport(
                importer="blue",
                imported="green",
                line_number=1,
                line_contents="from blue import green.one",
            ),
            DirectImport(
                importer="blue",
                imported="green",
                line_number=3,
                line_contents="from blue import green.two",
            ),
        ]) == {
            DirectImport(
                importer="blue",
                imported="green",
            ),
        }

    This is to make it easy for the calling function to remove the set of imports from a graph
    without attempting to remove certain imports twice.
    """
    imports_without_metadata = {
        DirectImport(imported=i.imported, importer=i.importer) for i in imports
    }
    # Why don't we return a set here? Because we want to preserve the order to make it
    # more deterministic.
    return sorted(imports_without_metadata, key=lambda di: (di.importer.name, di.imported.name))



import abc


class FileSystem(abc.ABC):
    """
    Abstraction around file system calls.
    """

    @abc.abstractmethod
    def join(self, *components: str) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def read(self, file_name: str) -> str:
        """
        Given a file name, return the contents of the file.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def exists(self, file_name: str) -> bool:
        """
        Return whether a file exists.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def getcwd(self) -> str:
        """
        Return the current working directory.
        """
        raise NotImplementedError



import abc
from typing import List, Optional

from grimp import ImportGraph


class GraphBuilder(abc.ABC):
    @abc.abstractmethod
    def build(
        self,
        root_package_names: List[str],
        cache_dir: Optional[str],
        include_external_packages: bool = False,
        exclude_type_checking_imports: bool = False,
    ) -> ImportGraph:
        raise NotImplementedError



from __future__ import annotations

from typing import Iterable, List, cast

from grimp import ImportGraph

from importlinter.application import contract_utils, output
from importlinter.application.contract_utils import AlertLevel
from importlinter.configuration import settings
from importlinter.domain import fields
from importlinter.domain.contract import Contract, ContractCheck
from importlinter.domain.helpers import module_expressions_to_modules
from importlinter.domain.imports import Module

from ._common import format_line_numbers


class ForbiddenContract(Contract):
    """
    Forbidden contracts check that one set of modules are not imported by another set of modules.
    Indirect imports will also be checked.
    Configuration options:
        - source_modules:    A list of Modules that should not import the forbidden modules.
        - forbidden_modules: A list of Modules that should not be imported by the source modules.
        - ignore_imports:    A set of ImportExpressions. These imports will be ignored if the import
                             would cause a contract to be broken, adding it to the set will cause
                             the contract be kept instead. (Optional.)
        - allow_indirect_imports:  Whether to allow indirect imports to forbidden modules.
                             "True" or "true" will be treated as True. (Optional.)
        - unmatched_ignore_imports_alerting: Decides how to report when the expression in the
                             `ignore_imports` set is not found in the graph. Valid values are
                             "none", "warn", "error". Default value is "error".
        - as_packages:       Whether to treat the source and forbidden modules as packages. If
                             False, each of the modules passed in will be treated as a module
                             rather than a package. Default behaviour is True (treat modules as
                             packages).
    """

    type_name = "forbidden"

    source_modules = fields.ListField(subfield=fields.ModuleExpressionField())
    forbidden_modules = fields.ListField(subfield=fields.ModuleExpressionField())
    ignore_imports = fields.SetField(subfield=fields.ImportExpressionField(), required=False)
    allow_indirect_imports = fields.BooleanField(required=False, default=False)
    unmatched_ignore_imports_alerting = fields.EnumField(AlertLevel, default=AlertLevel.ERROR)
    as_packages = fields.BooleanField(required=False, default=True)

    def check(self, graph: ImportGraph, verbose: bool) -> ContractCheck:
        is_kept = True
        invalid_chains = []

        warnings = contract_utils.remove_ignored_imports(
            graph=graph,
            ignore_imports=self.ignore_imports,  # type: ignore
            unmatched_alerting=self.unmatched_ignore_imports_alerting,  # type: ignore
        )

        source_modules = list(
            module_expressions_to_modules(
                graph,
                self.source_modules,  # type: ignore
            )
        )
        forbidden_modules = list(
            module_expressions_to_modules(
                graph,
                self.forbidden_modules,  # type: ignore
            )
        )

        self._check_all_modules_exist_in_graph(source_modules, graph)
        self._check_external_forbidden_modules(forbidden_modules)

        # We only need to check for illegal imports for forbidden modules that are in the graph.
        forbidden_modules_in_graph = [m for m in forbidden_modules if m.name in graph.modules]

        def sort_key(module):
            return module.name

        for source_module in sorted(source_modules, key=sort_key):
            for forbidden_module in sorted(forbidden_modules_in_graph, key=sort_key):
                output.verbose_print(
                    verbose,
                    "Searching for import chains from "
                    f"{source_module} to {forbidden_module}...",
                )
                with settings.TIMER as timer:
                    subpackage_chain_data = {
                        "upstream_module": forbidden_module.name,
                        "downstream_module": source_module.name,
                        "chains": [],
                    }

                    if str(self.allow_indirect_imports).lower() == "true":
                        chains = self._get_direct_chains(
                            source_module, forbidden_module, graph, self.as_packages  # type:ignore
                        )
                    else:
                        chains = graph.find_shortest_chains(
                            importer=source_module.name,
                            imported=forbidden_module.name,
                            as_packages=self.as_packages,  # type:ignore
                        )
                    if chains:
                        is_kept = False
                        for chain in sorted(chains):
                            chain_data = []
                            for importer, imported in [
                                (chain[i], chain[i + 1]) for i in range(len(chain) - 1)
                            ]:
                                import_details = graph.get_import_details(
                                    importer=importer, imported=imported
                                )
                                line_numbers = tuple(j["line_number"] for j in import_details)
                                chain_data.append(
                                    {
                                        "importer": importer,
                                        "imported": imported,
                                        "line_numbers": line_numbers,
                                    }
                                )
                            subpackage_chain_data["chains"].append(chain_data)  # type: ignore
                if subpackage_chain_data["chains"]:
                    invalid_chains.append(subpackage_chain_data)
                if verbose:
                    chain_count = len(subpackage_chain_data["chains"])
                    pluralized = "s" if chain_count != 1 else ""
                    output.print(
                        f"Found {chain_count} illegal chain{pluralized} "
                        f"in {timer.duration_in_s}s.",
                    )

        # Sorting by upstream and downstream module ensures that the output is deterministic
        # and that the same upstream and downstream modules are always adjacent in the output.
        def chain_sort_key(chain_data):
            return (chain_data["upstream_module"], chain_data["downstream_module"])

        return ContractCheck(
            kept=is_kept,
            warnings=warnings,
            metadata={"invalid_chains": sorted(invalid_chains, key=chain_sort_key)},
        )

    def render_broken_contract(self, check: "ContractCheck") -> None:
        count = 0
        for chains_data in check.metadata["invalid_chains"]:
            downstream, upstream = chains_data["downstream_module"], chains_data["upstream_module"]
            output.print_error(f"{downstream} is not allowed to import {upstream}:")
            output.new_line()
            count += len(chains_data["chains"])
            for chain in chains_data["chains"]:
                first_line = True
                for direct_import in chain:
                    importer, imported = direct_import["importer"], direct_import["imported"]
                    line_numbers = format_line_numbers(direct_import["line_numbers"])
                    import_string = f"{importer} -> {imported} ({line_numbers})"
                    if first_line:
                        output.print_error(f"-   {import_string}", bold=False)
                        first_line = False
                    else:
                        output.indent_cursor()
                        output.print_error(import_string, bold=False)
                output.new_line()

            output.new_line()

    def _check_all_modules_exist_in_graph(
        self, modules: Iterable[Module], graph: ImportGraph
    ) -> None:
        for module in modules:
            if module.name not in graph.modules:
                raise ValueError(f"Module '{module.name}' does not exist.")

    def _check_external_forbidden_modules(self, forbidden_modules) -> None:
        external_forbidden_modules = self._get_external_forbidden_modules(forbidden_modules)
        if external_forbidden_modules:
            if self._graph_was_built_with_externals():
                for module in external_forbidden_modules:
                    if module.root_package_name != module.name:
                        raise ValueError(
                            f"Invalid forbidden module {module}: "
                            "subpackages of external packages are not valid."
                        )
            else:
                raise ValueError(
                    "The top level configuration must have include_external_packages=True "
                    "when there are external forbidden modules."
                )

    def _get_external_forbidden_modules(self, forbidden_modules) -> set[Module]:
        root_packages = [Module(name) for name in self.session_options["root_packages"]]
        return {
            forbidden_module
            for forbidden_module in cast(List[Module], forbidden_modules)
            if not any(
                forbidden_module.is_in_package(root_package) for root_package in root_packages
            )
        }

    def _graph_was_built_with_externals(self) -> bool:
        return str(self.session_options.get("include_external_packages")).lower() == "true"

    def _get_direct_chains(
        self,
        source_package: Module,
        forbidden_package: Module,
        graph: ImportGraph,
        as_packages: bool,
    ) -> set[tuple[str, ...]]:
        chains: set[tuple[str, ...]] = set()
        source_modules = (
            self._get_all_modules_in_package(source_package, graph)
            if as_packages
            else {source_package}
        )
        forbidden_modules = (
            self._get_all_modules_in_package(forbidden_package, graph)
            if as_packages
            else {forbidden_package}
        )
        for source_module in source_modules:
            imported_module_names = graph.find_modules_directly_imported_by(source_module.name)
            for imported_module_name in imported_module_names:
                imported_module = Module(imported_module_name)
                if imported_module in forbidden_modules:
                    chains.add((source_module.name, imported_module.name))
        return chains

    def _get_all_modules_in_package(self, module: Module, graph: ImportGraph) -> set[Module]:
        """
        Return all the modules in the supplied module, including itself.

        If the module is squashed, it will be treated as a single module.
        """
        importer_modules = {module}
        if not graph.is_module_squashed(module.name):
            importer_modules |= {Module(m) for m in graph.find_descendants(module.name)}
        return importer_modules



import abc
from typing import Optional

from ..user_options import UserOptions


class UserOptionReader(abc.ABC):
    @abc.abstractmethod
    def read_options(self, config_filename: Optional[str] = None) -> Optional[UserOptions]:
        raise NotImplementedError



from typing import Dict, Iterator, List, Tuple

from importlinter.domain.contract import Contract, ContractCheck, InvalidContractOptions
from grimp import ImportGraph


class Reporter:
    ...


class ExceptionReporter:
    ...


class Report:
    def __init__(
        self, graph: ImportGraph, show_timings: bool, graph_building_duration: int
    ) -> None:
        self.graph = graph
        self.show_timings = show_timings
        self.graph_building_duration = graph_building_duration
        self.could_not_run = False
        self.invalid_contract_options: Dict[str, InvalidContractOptions] = {}
        self.contains_failures = False
        self.contracts: List[Contract] = []
        self._check_map: Dict[Contract, ContractCheck] = {}
        self._durations: Dict[Contract, int] = {}
        self.warnings_count = 0
        self.broken_count = 0
        self.kept_count = 0
        self.module_count = len(graph.modules)
        self.import_count = graph.count_imports()

    def add_contract_check(
        self, contract: Contract, contract_check: ContractCheck, duration: int
    ) -> None:
        self.contracts.append(contract)
        self._check_map[contract] = contract_check
        self._durations[contract] = duration
        self.warnings_count += len(contract_check.warnings)
        if contract_check.kept:
            self.kept_count += 1
        else:
            self.broken_count += 1
            self.contains_failures = True

    def get_contracts_and_checks(self) -> Iterator[Tuple[Contract, ContractCheck]]:
        for contract in self.contracts:
            yield contract, self._check_map[contract]

    def get_duration(self, contract) -> int:
        return self._durations[contract]

    def add_invalid_contract_options(
        self, contract_name: str, exception: InvalidContractOptions
    ) -> None:
        self.invalid_contract_options[contract_name] = exception
        self.could_not_run = True
        self.contains_failures = True



from typing import Optional

import click

from importlinter.application.ports.printing import Printer


class ClickPrinter(Printer):
    """
    Console printer that uses Click's formatting helpers.
    """

    def print(
        self, text: str = "", bold: bool = False, color: Optional[str] = None, newline: bool = True
    ) -> None:
        click.secho(text, bold=bold, fg=color, nl=newline)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Sequence, cast

import grimp
from typing_extensions import TypedDict

from importlinter.application import contract_utils, output
from importlinter.application.contract_utils import AlertLevel
from importlinter.domain import fields
from importlinter.domain.contract import Contract, ContractCheck, InvalidContractOptions
from importlinter.domain.helpers import module_expressions_to_modules
from importlinter.domain.imports import Module

from ._common import DetailedChain, build_detailed_chain_from_route, render_chain_data


_INDEPENDENT_LAYER_DELIMITER = "|"
_NON_INDEPENDENT_LAYER_DELIMITER = ":"


@dataclass(frozen=True)
class ModuleTail:
    """
    The "tail" of a module name.
    If no `containers` are specified then this is the full module name.
    However, if `containers` are specified then the full module name is made
    by combining the container name with this tail name.
    """

    name: str
    is_optional: bool = False


@dataclass(frozen=True)
class Layer:
    module_tails: set[ModuleTail]
    is_independent: bool = True


class LayerField(fields.Field):
    def parse(self, raw_data: str | list) -> Layer:
        raw_string = fields.StringField().parse(raw_data)
        raw_module_tails = self._parse_raw_module_tails(raw_string)
        module_tails = set()
        for raw_module_tail in raw_module_tails:
            if raw_module_tail.startswith("(") and raw_module_tail.endswith(")"):
                module_tail = raw_module_tail[1:-1]
                is_optional = True
            else:
                module_tail = raw_module_tail
                is_optional = False
            module_tails.add(ModuleTail(name=module_tail, is_optional=is_optional))
        is_independent = self._parse_is_independent(raw_string)
        return Layer(module_tails, is_independent=is_independent)

    @staticmethod
    def _parse_raw_module_tails(raw_string: str) -> list[str]:
        if _INDEPENDENT_LAYER_DELIMITER in raw_string:
            return [item.strip() for item in raw_string.split(_INDEPENDENT_LAYER_DELIMITER)]
        elif _NON_INDEPENDENT_LAYER_DELIMITER in raw_string:
            return [item.strip() for item in raw_string.split(_NON_INDEPENDENT_LAYER_DELIMITER)]
        else:
            return [raw_string.strip()]

    @staticmethod
    def _parse_is_independent(raw_string: str) -> bool:
        if (
            _INDEPENDENT_LAYER_DELIMITER in raw_string
            and _NON_INDEPENDENT_LAYER_DELIMITER in raw_string
        ):
            raise fields.ValidationError(
                "Layer cannot have a mixture of independent and non-independent elements."
            )
        elif _INDEPENDENT_LAYER_DELIMITER in raw_string:
            return True
        elif _NON_INDEPENDENT_LAYER_DELIMITER in raw_string:
            return False
        else:
            return True


class _LayerChainData(TypedDict):
    importer: str
    imported: str
    routes: list[DetailedChain]


_UNKNOWN_LINE_NUMBER = -1


class LayersContract(Contract):
    """
    Defines a 'layered architecture' where there is a unidirectional dependency flow.

    Specifically, higher layers may depend on lower layers, but not the other way around.
    To allow for a repeated pattern of layers across a project, you may also define a set of
    'containers', which are treated as the parent package of the layers.

    Layers are required by default: if a layer is listed in the contract, the contract will be
    broken if the layer doesn’t exist. You can make a layer optional by wrapping it in parentheses.

    Configuration options:

        - layers:             An ordered list of layers. Each layer is the name of a module relative
                              to its parent package. The order is from higher to lower level layers.
                              TODO docs.
        - containers:         A list of the parent Modules of the layers. (Optional.)
        - ignore_imports:     A set of ImportExpressions. These imports will be ignored: if the
                              import would cause a contract to be broken, adding it to the set will
                              cause the contract be kept instead. (Optional.)
        - unmatched_ignore_imports_alerting:
                              Decides how to report when the expression in the `ignore_imports` set
                              is not found in the graph. Valid values are "none", "warn", "error".
                              (Optional, defaults to "error".)
        - exhaustive:         If True, check that the contract declares every possible layer in
                              its list of layers to check. (Optional, default False.)
        - exhaustive_ignores: A set of potential layers to ignore in exhaustiveness checks.
                              (Optional.)
    """

    type_name = "layers"

    layers = fields.ListField(subfield=LayerField())
    containers = fields.ListField(subfield=fields.ModuleExpressionField(), required=False)
    ignore_imports = fields.SetField(subfield=fields.ImportExpressionField(), required=False)
    unmatched_ignore_imports_alerting = fields.EnumField(AlertLevel, default=AlertLevel.ERROR)
    exhaustive = fields.BooleanField(default=False)
    exhaustive_ignores = fields.SetField(subfield=fields.StringField(), required=False)

    def validate(self) -> None:
        if self.exhaustive and not self.containers:
            raise InvalidContractOptions(
                {
                    "exhaustive": (
                        "The exhaustive option is not supported for contracts without containers."
                    )
                }
            )

    def check(self, graph: grimp.ImportGraph, verbose: bool) -> ContractCheck:
        warnings = contract_utils.remove_ignored_imports(
            graph=graph,
            ignore_imports=self.ignore_imports,  # type: ignore
            unmatched_alerting=self.unmatched_ignore_imports_alerting,  # type: ignore
        )
        self.all_module_tails = self._get_all_module_tails_from_layers(self.layers)  # type: ignore

        containers = set()
        if self.containers:
            modules = module_expressions_to_modules(graph, self.containers)  # type: ignore
            containers = {m.name for m in modules}

        if self.containers:
            self._validate_containers(graph, containers)
        else:
            self._check_all_containerless_layers_exist(graph)

        undeclared_modules = self._get_undeclared_modules(graph, containers)

        dependencies = graph.find_illegal_dependencies_for_layers(
            layers=self._grimpify_layers(self.layers),  # type: ignore
            containers=containers,
        )
        invalid_chains = self._build_invalid_chains(dependencies, graph)

        return ContractCheck(
            kept=not (dependencies or undeclared_modules),
            warnings=warnings,
            metadata={
                "invalid_dependencies": invalid_chains,
                "undeclared_modules": undeclared_modules,
            },
        )

    def _get_all_module_tails_from_layers(self, layers: Sequence[Layer]) -> set[ModuleTail]:
        flattened = set()
        for layer in layers:
            flattened.update(layer.module_tails)
        return flattened

    def render_broken_contract(self, check: ContractCheck) -> None:
        for chains_data in cast(List[_LayerChainData], check.metadata["invalid_dependencies"]):
            higher_layer, lower_layer = (chains_data["imported"], chains_data["importer"])
            output.print(f"{lower_layer} is not allowed to import {higher_layer}:")
            output.new_line()

            for chain_data in chains_data["routes"]:
                render_chain_data(chain_data)
                output.new_line()

            output.new_line()

        if check.metadata["undeclared_modules"]:
            output.print("The following modules are not listed as layers:")
            output.new_line()
            for module in sorted(check.metadata["undeclared_modules"]):
                output.print_error(f"- {module}", bold=False)
            output.new_line()
            output.print(
                "(Since this contract is marked as 'exhaustive', every child of every "
                "container must be declared as a layer.)"
            )
            output.new_line()

    def _validate_containers(self, graph: grimp.ImportGraph, containers: set[str]) -> None:
        root_package_names = self.session_options["root_packages"]
        root_packages = tuple(Module(name) for name in root_package_names)

        for container in containers:
            if not any(
                Module(container).is_in_package(root_package) for root_package in root_packages
            ):
                if len(root_package_names) == 1:
                    root_package_name = root_package_names[0]
                    error_message = (
                        f"Invalid container '{container}': a container must either be a "
                        f"subpackage of {root_package_name}, or {root_package_name} itself."
                    )
                else:
                    packages_string = ", ".join(root_package_names)
                    error_message = (
                        f"Invalid container '{container}': a container must either be a root "
                        f"package, or a subpackage of one of them. "
                        f"(The root packages are: {packages_string}.)"
                    )
                raise ValueError(error_message)
            self._check_all_layers_exist_for_container(container, graph)

    def _check_all_layers_exist_for_container(
        self, container: str, graph: grimp.ImportGraph
    ) -> None:
        for module_tail in self.all_module_tails:
            if module_tail.is_optional:
                continue
            module_name = ".".join([container, module_tail.name])
            if module_name not in graph.modules:
                raise ValueError(
                    f"Missing layer in container '{container}': "
                    f"module {module_name} does not exist."
                )

    def _get_undeclared_modules(self, graph: grimp.ImportGraph, containers: set[str]) -> set[str]:
        if not self.exhaustive:
            return set()

        undeclared_modules = set()

        exhaustive_ignores: set[str] = self.exhaustive_ignores or set()  # type: ignore
        module_tail_names: set[str] = {module_tail.name for module_tail in self.all_module_tails}
        declared_module_tails = module_tail_names | exhaustive_ignores

        for container in containers:
            for module in graph.find_children(container):
                module_tail = module.rpartition(".")[-1]
                if module_tail not in declared_module_tails:
                    undeclared_modules.add(f"{container}.{module_tail}")

        return undeclared_modules

    def _check_all_containerless_layers_exist(self, graph: grimp.ImportGraph) -> None:
        for layer in self.layers:  # type: ignore
            for module_tail in layer.module_tails:
                if module_tail.is_optional:
                    continue
                if module_tail.name not in graph.modules:
                    raise ValueError(
                        f"Missing layer '{module_tail.name}': "
                        f"module {module_tail.name} does not exist."
                    )

    def _module_from_module_tail(
        self, module_tail: ModuleTail, container: str | None = None
    ) -> Module:
        if container:
            name = ".".join([container, module_tail.name])
        else:
            name = module_tail.name
        return Module(name)

    def _build_invalid_chains(
        self, dependencies: set[grimp.PackageDependency], graph: grimp.ImportGraph
    ) -> list[_LayerChainData]:
        return [
            {
                "imported": dependency.imported,
                "importer": dependency.importer,
                "routes": [build_detailed_chain_from_route(c, graph) for c in dependency.routes],
            }
            for dependency in dependencies
        ]

    @staticmethod
    def _grimpify_layers(layers: list[Layer]) -> list[grimp.Layer]:
        return [
            grimp.Layer(
                *{module_tail.name for module_tail in layer.module_tails},
                independent=layer.is_independent,
            )
            for layer in layers
        ]



from typing import List, Optional

import grimp
from grimp import ImportGraph

from importlinter.application.ports import building as ports


class GraphBuilder(ports.GraphBuilder):
    """
    GraphBuilder that just uses Grimp's standard build_graph function.
    """

    def build(
        self,
        root_package_names: List[str],
        cache_dir: Optional[str],
        include_external_packages: bool = False,
        exclude_type_checking_imports: bool = False,
    ) -> ImportGraph:
        return grimp.build_graph(
            *root_package_names,
            include_external_packages=include_external_packages,
            exclude_type_checking_imports=exclude_type_checking_imports,
            cache_dir=cache_dir,
        )



import enum
from typing import List, Optional, Sequence, Set


from importlinter.domain.helpers import MissingImport
from importlinter.domain.imports import ImportExpression, DirectImport, Module
from grimp import ImportGraph


class AlertLevel(enum.Enum):
    NONE = "none"
    WARN = "warn"
    ERROR = "error"


def remove_ignored_imports(
    graph: ImportGraph,
    ignore_imports: Optional[Sequence[ImportExpression]],
    unmatched_alerting: AlertLevel,
) -> List[str]:
    """
    Remove any ignored imports from the graph.

    Args:
        graph:              The graph that is being checked by a contract.
        ignore_imports:     Any import expressions that indicate imports to ignore.
        unmatched_alerting: An AlertLevel that indicates how to handle any import expressions that
                            don't match any imports. AlertLevel.NONE will ignore them,
                            AlertLevel.WARN will warn for each one, and AlertLevel.ERROR will raise
                            a MissingImport for the first one encountered.

    Returns:
        A list of any warnings to be surfaced to the user.
    """
    imports_to_remove = set()
    unresolved_expressions = set()
    for import_expression in ignore_imports or []:
        matched_imports = graph.find_matching_direct_imports(
            import_expression=str(import_expression)
        )
        if matched_imports:
            imports_to_remove.update(
                {
                    DirectImport(
                        importer=Module(matched_import["importer"]),
                        imported=Module(matched_import["imported"]),
                    )
                    for matched_import in matched_imports
                }
            )
        else:
            unresolved_expressions.add(import_expression)

    warnings = _handle_unresolved_import_expressions(
        unresolved_expressions,
        unmatched_alerting,
    )

    for import_to_remove in imports_to_remove:
        graph.remove_import(
            importer=import_to_remove.importer.name,
            imported=import_to_remove.imported.name,
        )

    return warnings


# Private functions
# -----------------


def _handle_unresolved_import_expressions(
    expressions: Set[ImportExpression], alert_level: AlertLevel
) -> List[str]:
    """
    Handle any unresolved import expressions based on the supplied alert level.

    Intended to be called while checking a contract.

    Returns:
        A list of any warnings to be surfaced to the user.
    """
    if alert_level is AlertLevel.NONE:
        return []
    if not expressions:
        return []

    if alert_level is AlertLevel.WARN:
        return [_build_missing_import_message(expression) for expression in expressions]
    else:  # AlertLevel.ERROR
        first_expression = sorted(expressions, key=lambda expression: str(expression))[0]
        raise MissingImport(_build_missing_import_message(first_expression))


def _build_missing_import_message(expression: ImportExpression) -> str:
    return f"No matches for ignored import {expression}."



import abc
from enum import Enum
from typing import Generic, Iterable, List, Set, Type, TypeVar, Union, cast

from importlinter.domain.imports import ImportExpression, Module, ModuleExpression

FieldValue = TypeVar("FieldValue")


class NotSupplied:
    """Sentinel to use in place of None for a default argument value."""

    pass


class ValidationError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message


class Field(Generic[FieldValue], abc.ABC):
    """
    Base class for containers for some data on a Contract.

    Arguments:
        - required: if the field requires a value (default True).
        - default: a value to use if no value is supplied.

    Designed to be subclassed, Fields should override the ``parse`` method.
    """

    def __init__(
        self,
        required: Union[bool, Type[NotSupplied]] = NotSupplied,
        default: Union[FieldValue, Type[NotSupplied]] = NotSupplied,
    ) -> None:
        if default is NotSupplied:
            if required is NotSupplied:
                self.required = True
            else:
                required = cast(bool, required)
                self.required = required
        else:
            # A default was supplied.
            if required is True:
                # It doesn't make sense to require a field and provide a default.
                raise ValueError("A required field cannot also provide a default value.")
            else:
                self.required = False

        self.default = default

    @abc.abstractmethod
    def parse(self, raw_data: Union[str, List[str]]) -> FieldValue:
        """
        Given some raw data supplied by a user, return some clean data.

        Raises:
            ValidationError if the data is invalid.
        """
        raise NotImplementedError


class StringField(Field):
    """
    A field for single values of strings.
    """

    def parse(self, raw_data: Union[str, List]) -> str:
        if isinstance(raw_data, list):
            raise ValidationError("Expected a single value, got multiple values.")
        return str(raw_data)


class BooleanField(Field):
    """
    A field for single values of booleans.
    """

    def parse(self, raw_data: Union[str, List]) -> bool:
        if isinstance(raw_data, list):
            raise ValidationError("Expected a single value, got multiple values.")

        if raw_data.lower() == "true":
            return True
        elif raw_data.lower() == "false":
            return False
        else:
            raise ValidationError(f"Could not parse a boolean from '{raw_data}'.")


class BaseMultipleValueField(Field):
    """
    An abstract field for multiple values of any type.

    Arguments:
        - subfield: An instance of a single-value Field. Each item in the iterable will be
                    the return value of this subfield.

    """

    def __init__(self, subfield: Field, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.subfield = subfield

    @abc.abstractmethod
    def parse(self, raw_data: Union[str, List]) -> Iterable[FieldValue]:
        if isinstance(raw_data, tuple):
            raw_data = list(raw_data)
        if not isinstance(raw_data, list):
            raw_data = [raw_data]  # Single values should just be treated as a single item list.
        clean_list = []
        for raw_line in raw_data:
            # Ignore blank lines
            if not raw_line.strip():
                continue
            clean_list.append(self.subfield.parse(raw_line))
        return clean_list


class ListField(BaseMultipleValueField):
    """
    A field for multiple values of any type.

    Fields values are returned in list sorted by parsing order.

    Usage:

        field = ListField(subfield=AnotherField())
    """

    def parse(self, raw_data: Union[str, List]) -> List[FieldValue]:
        return list(super().parse(raw_data))


class SetField(BaseMultipleValueField):
    """
    A field for multiple, unique values of any type.

    Fields values are returned inordered in set.

    Usage:

        field = SetField(subfield=AnotherField())

    """

    def parse(self, raw_data: Union[str, List]) -> Set[FieldValue]:
        return set(super().parse(raw_data))


class ModuleField(Field):
    """
    A field for Modules.
    """

    def parse(self, raw_data: Union[str, List]) -> Module:
        return Module(StringField().parse(raw_data))


class ModuleExpressionField(Field):
    """
    A field for ModuleExpressions.

    Accepts strings in the form:
        "mypackage.foo.importer"
        "mypackage.foo.*"
        "mypackage.*.importer"
        "mypackage.**"
    """

    def parse(self, expression: Union[str, List[str]]) -> ModuleExpression:
        if isinstance(expression, list):
            raise ValidationError("Expected a single value, got multiple values.")

        last_wildcard = None
        for part in expression.split("."):
            if "**" == last_wildcard and ("*" == part or "**" == part):
                raise ValidationError("A recursive wildcard cannot be followed by a wildcard.")
            if "*" == last_wildcard and "**" == part:
                raise ValidationError("A wildcard cannot be followed by a recursive wildcard.")
            if "*" == part or "**" == part:
                last_wildcard = part
                continue
            if "*" in part:
                raise ValidationError("A wildcard can only replace a whole module.")
            last_wildcard = None

        return ModuleExpression(expression)


class ImportExpressionField(Field):
    """
    A field for ImportExpressions.

    Expects raw data in the form:
        "mypackage.foo.importer -> mypackage.bar.imported".

    In addition, it handles wildcards:
        "mypackage.*.importer -> mypackage.bar.*"
        "mypackage.**.importer -> mypackage.bar.**"
    """

    def parse(self, raw_data: Union[str, List]) -> ImportExpression:
        string = StringField().parse(raw_data)
        importer, _, imported = string.partition("->")
        # Remove any whitespace around the module string
        importer = importer.strip()
        imported = imported.strip()

        if not (importer and imported):
            raise ValidationError('Must be in the form "package.importer -> package.imported".')

        return ImportExpression(
            importer=ModuleExpressionField().parse(importer),
            imported=ModuleExpressionField().parse(imported),
        )


class EnumField(Field):
    """
    A field constrained by the values of an Enum.

    Currently only Enums with string values are supported.

    Raises a ValidationError if the supplied string value does not match one of the Enum
    members' values.

    Example:

        import enum

        class Color(enum.Enum):
            RED = "red"
            BLUE = "blue"
            LIGHT_BLUE = "light blue"

        field = EnumField(Color, default=Color.RED)

        assert field.parse("blue") == Color.BLUE
        assert field.parse("light blue") == Color.LIGHT_BLUE
        assert field.parse("") == Color.RED
    """

    def __init__(self, enum: Type[Enum], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._check_supported_enum_class(enum)
        self.enum = enum

    def parse(self, raw_data: Union[str, List]) -> Enum:
        if isinstance(raw_data, list):
            raise ValidationError("Expected a single value, got multiple values.")

        stripped_data = raw_data.strip()

        if stripped_data == "":
            return cast(Enum, self.default)

        member_by_value = {m.value: m for m in self.enum}
        try:
            return member_by_value[stripped_data]
        except KeyError:
            values = list(member_by_value.keys())
            expectation_string = ", ".join(f"'{i}'" for i in values[:-1]) + f" or '{values[-1]}'"
            raise ValidationError(
                f"Invalid value '{stripped_data}': " f"expected {expectation_string}."
            )

    def _check_supported_enum_class(self, enum: Type[Enum]) -> None:
        for member in enum:
            # Check it's a string.
            if not isinstance(member.value, str):
                raise TypeError(
                    "Unsupported Enum for EnumField: member values must all be strings."
                )



import os
import sys
from logging import config as logging_config
from typing import Optional, Tuple, Type, Union

import click

from importlinter.application.sentinels import NotSupplied

from . import configuration
from .application import use_cases

configuration.configure()

EXIT_STATUS_SUCCESS = 0
EXIT_STATUS_ERROR = 1


@click.command()
@click.option("--config", default=None, help="The config file to use.")
@click.option(
    "--contract",
    default=list,
    multiple=True,
    help="Limit the check to the supplied contract identifier. May be passed multiple times.",
)
@click.option("--cache-dir", default=None, help="The directory to use for caching.")
@click.option("--no-cache", is_flag=True, help="Disable caching.")
@click.option("--debug", is_flag=True, help="Run in debug mode.")
@click.option(
    "--show-timings",
    is_flag=True,
    help="Show times taken to build the graph and to check each contract.",
)
@click.option(
    "--verbose",
    is_flag=True,
    help="Noisily output progress as we go along.",
)
def lint_imports_command(
    config: Optional[str],
    contract: Tuple[str, ...],
    cache_dir: Optional[str],
    no_cache: bool,
    debug: bool,
    show_timings: bool,
    verbose: bool,
) -> int:
    """
    Check that a project adheres to a set of contracts.
    """
    exit_code = lint_imports(
        config_filename=config,
        limit_to_contracts=contract,
        cache_dir=cache_dir,
        no_cache=no_cache,
        is_debug_mode=debug,
        show_timings=show_timings,
        verbose=verbose,
    )
    sys.exit(exit_code)


def lint_imports(
    config_filename: Optional[str] = None,
    limit_to_contracts: Tuple[str, ...] = (),
    cache_dir: Optional[str] = None,
    no_cache: bool = False,
    is_debug_mode: bool = False,
    show_timings: bool = False,
    verbose: bool = False,
) -> int:
    """
    Check that a project adheres to a set of contracts.

    This is the main function that runs the linter.

    Args:
        config_filename:    the filename to use to parse user options.
        limit_to_contracts: if supplied, only lint the contracts with the supplied ids.
        cache_dir:          the directory to use for caching, defaults to '.import_linter_cache'.
        no_cache:           if True, disable caching.
        is_debug_mode:      whether debugging should be turned on. In debug mode, exceptions are
                            not swallowed at the top level, so the stack trace can be seen.
        show_timings:       whether to show the times taken to build the graph and to check
                            each contract.
        verbose:            if True, noisily output progress as it goes along.

    Returns:
        EXIT_STATUS_SUCCESS or EXIT_STATUS_ERROR.
    """
    # Add current directory to the path, as this doesn't happen automatically.
    sys.path.insert(0, os.getcwd())

    _configure_logging(verbose)

    combined_cache_dir = _combine_caching_arguments(cache_dir, no_cache)

    passed = use_cases.lint_imports(
        config_filename=config_filename,
        limit_to_contracts=limit_to_contracts,
        cache_dir=combined_cache_dir,
        is_debug_mode=is_debug_mode,
        show_timings=show_timings,
        verbose=verbose,
    )

    if passed:
        return EXIT_STATUS_SUCCESS
    else:
        return EXIT_STATUS_ERROR


def _combine_caching_arguments(
    cache_dir: Optional[str], no_cache: bool
) -> Union[str, None, Type[NotSupplied]]:
    if no_cache:
        return None
    if cache_dir is None:
        return NotSupplied
    return cache_dir


def _configure_logging(verbose: bool) -> None:
    logger_names = ("importlinter", "grimp", "_rustgrimp")
    logging_config.dictConfig(
        {
            "version": 1,
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "level": "INFO" if verbose else "WARNING",
                    "stream": "ext://sys.stdout",
                },
            },
            "loggers": {
                logger_name: {
                    "level": "INFO",
                    "handlers": ["console"],
                }
                for logger_name in logger_names
            },
        }
    )
